package Gear::Remotes::GitUrl;

use strict;
use warnings;

our $verbose=0;

# for reference

# https://docs.gitlab.com/ee/user/project/pages/getting_started_part_one.html

# pubic REST API :
# https://docs.github.com/en/rest/reference/repos#get-a-repository
# В нём возвращается тип репозитория.

my $deadpatterns = {
  # gitorious is dead
  'gitorious' => [
      # ban anything on gitorious.org, just in case
    { 'pattern'     => qr'^https://(?:[^\./]+\.)?gitorious.org/',
      'description' => 'https://(…[-.])gitorious.org/… is dead'},
      ],
  #https://src.fedoraproject.org/cgit/rpms/
  # cgit no more
  'src.fedoraproject.org' => [
    { 'pattern'     => qr'^https://[^/]+/cgit/rpms/',
      'description' => 'https://(…[-.])[-.]…/cgit/rpms/… is dead'},
      ],

};

my $banpatterns = {
  'gitlab' => [
      # group, not a repo
    { 'pattern'     => qr'^https://[^/]+/groups/[^/#?]+$',
      'description' => 'https://(…[-.])gitlab[-.]…/groups/group'},
      ],
  'github' => [
    { 'pattern'     => qr'^https://[^/]+/[^/]+/[^/#?]+/wiki$',
      'description' => 'https://(…[-.])github[-.]…/owner/repo/wiki'},
    { 'pattern'     => qr'^https://[^\./]+\.github\.[^/]+/[^/#?]+/wiki$',
      'description' => 'https://owner.github[-.]…/repo/wiki'},
      ],
};

my $urlpatterns = {
  'gitlab' => [
    { 'pattern'     => qr'^https://[^/]+/[^/]+/[^/#?]+$',
     'description' => 'https://(…[-.])gitlab[-.]…/owner/repo'},
    { 'pattern'     => qr'^https://[^/]+/[^/]+(?:/[^/]+)+/[^/#?]+$',
     'description' => 'https://(…[-.])gitlab[-.]…/owner/subgroup../repo'},
      ],
  'pagure' => [
    { 'pattern'     => qr'^https://[^/]+/[^/#?]+$',
      'description' => 'https://pagure.io/repo'},
    { 'pattern'     => qr'^https://[^/]+/[^/]+/[^/#?]+$',
      'description' => 'https://pagure.io/namespace/repo'},
    { 'pattern'     => qr'^https://[^/]+/fork/[^/]+/[^/#?]+',
      'description' => 'https://pagure.io/fork/owner/repo'},
    { 'pattern'     => qr'^https://[^/]+/fork/[^/]+/[^/]+/[^/#?]+$',
      'description' => 'https://pagure.io/fork/owner/namespace/repo'},
      ],
  'gitea.com' => [{
      'pattern'     => qr'^https://[^/]+/[^/]+/[^/#?]+$',
      'description' => 'https://gitea.com/owner/repo'}],
  'github' => [
    { 'pattern'     => qr'^https://[^/]+/[^/]+/[^/#?]+$',
      'description' => 'https://(…[-.])github[-.]…/owner/repo'},
    { 'pattern'     => qr'^https://[^\./]+\.github\.[^/]+/[^/#?]+$',
      'description' => 'https://owner.github[-.]…/repo'},
      ],
  'code.googlesource.com' => [{
      'pattern'     => qr'^https://code\.googlesource\.com/[^#?]*[^/#?]+$',
      'description' => 'https://code.googlesource.com/…/repo'}],
  'googlesource' => [{
      'pattern'     => qr'^https://[^/.]+\.googlesource\.com/[^#?]*[^/#?]+$',
      'description' => 'https://(…[-.]).googlesource.com/…/repo'}],
  'bitbucket.org' => [{
      'pattern'     => qr'^https://[^/]+/[^/]+/[^/#?]+$',
      'description' => 'https://bitbucket.org/owner/repo'}],
  'git.code.sf.net' => [{
      'pattern'     => qr'^https://git\.code\.sf\.net/[^/]+/[^/]+/[^/#?]+$',
      'description' => 'https://git.code.sf.net/p/gprolog/code'}],
#  # gitorious is dead :(
#  'gitorious' => [
#    { 'pattern'     => qr'^https://(?:[^\./]+\.)?gitorious.org/[^/]+/[^/#?]+$',
#      'description' => 'https://(…[-.])gitorious.org/owner/repo'},
#      ],
};

sub normalize_url {
    # git:// -> https:// for github:
    # The unauthenticated git protocol on port 9418 is no longer supported.
    # Please see https://github.blog/2021-09-01-improving-git-protocol-security-github/ for more information.
    return $_[0]
	=~ s!^https?://!https://!r
	=~ s,/$,,r
	=~ s,^git(://(?:[^\./]+\.)?github),https$1,r
	;
}

sub warn_git_like_url {
    local $_ = $_[0];
    # contains raw space. not html quoted. not supported yet.
    if (/\s/) {
	return 'url contains raw space. not html quoted. not supported yet.';
    };
    if (m!^git://(?:[\w\d]+\.)?github\.!) {
	return 'github supports public git:// no more. Switch to [git+]https://';
    };

    my $normalized_url=&normalize_url($_);
    s!^(?:git|git\+https?|https?)://!!;
    return 'looks like an url to a file or an HTML page' if /\.(?:html?|md|zip|tar(?:\.(?:gz|bz2|xz|lz|lzma))?)$/;
    return 'git.altlinux.org - watch themselves?' if /^git.altlinux.org/;

    # ?&p=... and so on
    return '?&p=... and so on... not a git-like url' if m!^[^/]+\.[^/]+/.*[&?=]!;

    my $prefix;
    if (m!^([^\./]+)\.!) {
	$prefix=$1;
	return $prefix if $prefix eq 'hg' or $prefix eq 'svn' or $prefix eq 'cvs';
    }
    # full.domain.name
    if (m!^((?:[^\./]+\.)+[^\./]+)/!) {
	my $domainkey=$1;
	my $warn_git_host = &__warn_git_host($domainkey, $normalized_url);
	return $warn_git_host if defined $warn_git_host;
    }
    # middle domain part
    if (m!^(?:[^\./]+\.)?([^\./]+)\.[^\./]+/!) {
	my $domainkey=$1;
	my $warn_git_host = &__warn_git_host($domainkey, $normalized_url);
	return $warn_git_host if defined $warn_git_host;
    }
    # last domain part
    if (m!^([^\./]+)(?:\.[^\./]+)*\.[^\./]+/!) {
	my $domainkey=$1;
	my $warn_git_host = &__warn_git_host($domainkey, $normalized_url);
	return $warn_git_host if defined $warn_git_host;
    }

    return;
}

sub __warn_git_host {
    my ($key, $url) = @_;
    my $deadpatterns = $deadpatterns->{$key};
    if ($deadpatterns) {
	foreach my $set (@$deadpatterns) {
	    if ($url =~ $set->{'pattern'}) {
		return $set->{'description'};
	    }
	}
    }
    return;
}

sub is_git_like_url {
    local $_ = $_[0];
    # no need to watch themselves
    my $ban_alt=1;
    return 0 if $ban_alt && m!^git://git.altlinux.org!;
    return 1 if m!^git://!;
    return 1 if m!^git\+https?://!;
    return 0 if not m!^https?://!;
    # contains raw space. not html quoted. not supported yet.
    return 0 if /\s/;
    my $normalized_url=&normalize_url($_);
    s!^https?://!!;
    return 0 if /\.(?:html?|md|zip|tar(?:\.(?:gz|bz2|xz|lz|lzma))?)$/;
    return 0 if $ban_alt && /^git.altlinux.org/;

    # ?&p=... and so on
    return 0 if m!^[^/]+\.[^/]+/.*[&?=]!;

    my $prefix;
    if (m!^([^\./]+)\.!) {
	$prefix=$1;
	return 0 if $prefix eq 'hg' or $prefix eq 'svn' or $prefix eq 'cvs';
    }
    # full.domain.name
    if (m!^((?:[^\./]+\.)+[^\./]+)/!) {
	my $domainkey=$1;
	my $is_git_host = &__is_git_host($domainkey, $normalized_url);
	return $is_git_host if defined $is_git_host;
    }
    # middle domain part
    if (m!^(?:[^\./]+\.)?([^\./]+)\.[^\./]+/!) {
	my $domainkey=$1;
	my $is_git_host = &__is_git_host($domainkey, $normalized_url);
	return $is_git_host if defined $is_git_host;
    }
    # last domain part
    if (m!^([^\./]+)(?:\.[^\./]+)*\.[^\./]+/!) {
	my $domainkey=$1;
	#warn "last: $domainkey\n";
	my $is_git_host = &__is_git_host($domainkey, $normalized_url);
	return $is_git_host if defined $is_git_host;
    }

    return 1 if $prefix and ($prefix eq 'cgit' or $prefix eq 'anongit' or $prefix eq 'git');
    return 1 if m!\.git/?$!;

    return;
}

sub __is_git_host {
    my ($key, $url) = @_;
    my $deadpatterns = $deadpatterns->{$key};
    if ($deadpatterns) {
	foreach my $set (@$deadpatterns) {
	    if ($url =~ $set->{'pattern'}) {
		warn "rejected: the URL: ", $url, ' match pattern for ',$key,' :',$set->{'description'},"\n" if $verbose>1;
		return 0;
	    }
	}
    }
    my $banpatterns = $banpatterns->{$key};
    if ($banpatterns) {
	foreach my $set (@$banpatterns) {
	    if ($url =~ $set->{'pattern'}) {
		warn "rejected: the URL: ", $url, ' match pattern for ',$key,' :',$set->{'description'},"\n" if $verbose>1;
		return 0;
	    }
	}
    }
    my $patterns = $urlpatterns->{$key};
    return if ! $patterns;
    foreach my $set (@$patterns) {
	if ($url =~ $set->{'pattern'}) {
	    warn "accepted: the URL: ", $url, ' match pattern for ',$key,' :',$set->{'description'},"\n" if $verbose>1;
	    return 1;
	}
    }
    warn "the URL: ", $url, ' does not match known git https://url patterns for ',$key,"\n" if $verbose;
    #return 1;
    return;
}

# process the raw value of rpm tags %{VCS},%{URL}
sub emptify_no_url {
    my $arg = shift;
    if (!defined($arg) or $arg eq '(none)') {
	return '';
    } elsif ($arg eq '0' or $arg =~ /^[Nn][Oo]/
	or $arg =~ /^[Oo][Ff][Ff]/
	or $arg =~ /^[Ff][Aa][Ll][Ss][Ee]/) {
	return '0';
    }
    return $arg;
}

1;
